Screen {
    align: center middle;
}

Input {
    width: auto;
    margin-bottom: 1;
}

Button {
    width: 16;
}

Container {
    layout: horizontal;
    width: 100%;
}

ListView {
    width: 100%;
    height: 25%;
}

Label {
    width: auto;
    margin-top: 1;
}