#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2024 Lanzhou University
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from liscopelens.utils import SPDXParser
from liscopelens.api import (
    BaseCompatiblityParser,
    BaseExceptionParser,
    BasePropagateParser,
    LicenseMap,
    ParserInitArgs,
    ScancodeExecutionError,
    ScancodeParser,
    check_compatibility,
    detect_license,
    detect_license_chunked,
)

__all__ = [
    "SPDXParser",
    "BaseCompatiblityParser",
    "BaseExceptionParser",
    "BasePropagateParser",
    "ParserInitArgs",
    "ScancodeParser",
    "check_compatibility",
    "LicenseMap",
    "detect_license",
    "detect_license_chunked",
    "ScancodeExecutionError",
]
